from jwt_utils.backend import JwtDecode as BaseDecode
from jwt_utils.backend import JwtEncode as BaseEncode
from jwt_utils.django.settings import api_settings


class JwtEncode(BaseEncode):
    signing_key = api_settings.SIGNING_KEY
    algorithm = api_settings.ALGORITHM
    iss = api_settings.ISSUER
    sub = api_settings.SUBJECT
    aud = api_settings.AUDIENCE
    exp = api_settings.EXPIRATION
    nbf = api_settings.NOT_BEFORE
    iat = api_settings.IAT
    auto_iat = api_settings.AUTO_IAT
    jti = api_settings.JTI
    auto_jti = api_settings.AUTO_JTI


class JwtDecode(BaseDecode):
    signing_key = api_settings.SIGNING_KEY
    algorithm = api_settings.ALGORITHM
    iss = api_settings.ISSUER
    aud = api_settings.AUDIENCE
    verify_iss = api_settings.VERIFY_ISSUER
    verify_aud = api_settings.VERIFY_AUDIENCE
    verify_exp = api_settings.VERIFY_EXPIRATION
    verify_nbf = api_settings.VERIFY_NOT_BEFORE
    verify_iat = api_settings.VERIFY_IAT



from calendar import timegm
from datetime import datetime, timedelta
from typing import List, Union
from uuid import uuid4

from jwt import InvalidAlgorithmError, InvalidTokenError
from jwt import decode as jwt_decode
from jwt import encode as jwt_encode


class BaseJwtToken:
    signing_key = None
    algorithm = 'HS256'
    iss = None
    aud = None

    def __init__(self, signing_key=None, algorithm='HS256', iss: str = None, aud: Union[str, List[str]] = None):
        """
            signing_key: str -> key used to encoder, ex: "1234"
            algorithm: str -> algorithm used to encode, exs: 'HS256', 'HS384', 'HS512', 'RS256', 'RS384', 'RS512'
            iss: str -> set token issuer
            aud: str | [str] -> set token subject
        """
        self.signing_key, self.algorithm = (signing_key or self.signing_key), (algorithm or self.algorithm)
        self.iss, self.aud = (iss or self.iss), (aud or self.aud)

        assert isinstance(self.signing_key, str) and self.signing_key is not None, "`signing_key` is required"
        assert self.algorithm in ['HS256', 'HS384', 'HS512', 'RS256', 'RS384', 'RS512']


class JwtEncode(BaseJwtToken):
    payload = {}
    sub = None
    exp = None
    nbf = None
    auto_iat = None
    iat = None
    auto_jti = None
    jti = None
    from_time = None

    def __init__(self, payload: dict = None, sub: str = None, exp: timedelta = None, nbf: timedelta = None,
                 auto_iat: bool = False, iat: datetime = None, auto_jti: bool = False, jti: str = None,
                 from_time: datetime = None, **kwargs):
        """
            payload: dict -> custom data, ex: {}
            sub: str -> set token subject
            exp: timedelta -> added to `from_time` to get token lifetime, ex: timedelta(minutes=5)
            nbf: timedelta -> added to `from_time` to get token invalid begore time, ex: timedelta(minutes=5)
            auto_iat: bool -> add `iat` with value `datetime.utcnow()`
            iat: datetime -> set issued token time
            auto_jti: bool -> add `jti` with `uuid4().hex`
            jti: str -> set token unique identifier
            from_time: datetime -> base datetime to calc `exp` and `nbf`, must be utc, ex: datetime.utcnow() (default)
            kwargs:
                signing_key: str -> key used to encoder, ex: "1234"
                algorithm: str -> algorithm used to encode, exs: 'HS256', 'HS384', 'HS512', 'RS256', 'RS384', 'RS512'
                iss: str -> set token issuer
                aud: str | [str] -> set token subject
        """
        super().__init__(**kwargs)
        self.payload = payload or self.payload
        self.from_time = from_time
        self.sub, self.exp, nbf = (sub or self.sub), (exp or self.exp), (nbf or self.nbf)
        self.auto_iat, self.iat = (auto_iat or self.auto_iat), (iat or self.iat),
        self.auto_jti, self.jti = (auto_jti or self.auto_jti), (jti or self.jti)
        self.from_time = (from_time or self.from_time)

    def set_claim(self, field, value=None):
        if value:
            self.payload[field] = value

    def mount_claims(self):
        self.payload, base_time = self.payload, self.from_time or datetime.utcnow()
        self.set_claim(field='iss', value=self.iss)
        self.set_claim(field='sub', value=self.sub)
        self.set_claim(field='aud', value=self.aud)

        if self.jti or self.auto_jti:
            jti = uuid4().hex if self.auto_jti and not self.jti else self.jti
            self.set_claim(field='jti', value=jti)

        if self.iat or self.auto_iat:
            iat = base_time if self.auto_iat and not self.iat else self.iat
            self.set_claim(field='iat', value=timegm(iat.utctimetuple()))

        if self.exp:
            exp = base_time + self.exp
            self.set_claim(field='exp', value=timegm(exp.utctimetuple()))

        if self.nbf:
            nbf = base_time + self.nbf
            self.set_claim(field='nbf', value=timegm(nbf.utctimetuple()))

    @staticmethod
    def mount_headers():
        return {}

    def encode(self):
        self.mount_claims()
        return jwt_encode(self.payload, self.signing_key, algorithm=self.algorithm, headers=self.mount_headers())


class JwtDecode(BaseJwtToken):
    verify_iss = False
    verify_aud = False
    verify_exp = False
    verify_nbf = False,
    verify_iat = False
    leeway = 0

    def __init__(self, token: str = None, verify_iss=False, verify_aud=False, verify_exp=False, verify_nbf=False,
                 verify_iat=False, leeway=0, **kwargs):
        """
            token: str -> valid jwt token, ex: ""
            verify_iss: bool -> verify jwt `issuer`
            verify_aud: bool -> verify jwt `audience`
            verify_exp: bool -> verify jwt `expiration`
            verify_nbf: bool -> verify jwt `not before time`
            verify_iat: bool -> verify jwt `issued at`
            leeway: int -> time margin for validation, ex: 30
            kwargs:
                signing_key: str -> key used to encoder, ex: "1234"
                algorithm: str -> algorithm used to encode, exs: 'HS256', 'HS384', 'HS512', 'RS256', 'RS384', 'RS512'
                iss: str -> set token issuer
                aud: str | [str] -> set token subject
        """
        super().__init__(**kwargs)
        self.token = token
        self.verify_iss, self.verify_aud = (verify_iss or self.verify_iss), (verify_aud or self.verify_aud)
        self.verify_exp, self.verify_nbf = (verify_exp or self.verify_exp), (verify_nbf or self.verify_nbf)
        self.verify_iat, self.leeway = (verify_iat or self.verify_iat), (leeway or self.leeway)

    def decode_raw(self):
        return jwt_decode(
            self.token,
            self.signing_key,
            algorithms=[self.algorithm] if not isinstance(self.algorithm, list) else self.algorithm,
            audience=self.aud,
            issuer=self.iss,
            leeway=self.leeway,
            options={
                'verify_iss': self.verify_iss,
                'verify_aud': self.verify_aud,
                'verify_exp': self.verify_exp,
                'verify_nbf': self.verify_nbf,
                'verify_iat': self.verify_iat,
                'verify_signature': True,
            },
        )

    def decode(self, raise_except=False):
        if raise_except:
            return self.decode_raw()
        
        else:
            try:
                return self.decode_raw()
            except InvalidAlgorithmError:
                return None
            except InvalidTokenError:
                return None



from django.conf import settings
from django.test.signals import setting_changed
from django.utils.translation import gettext_lazy as _

from rest_framework.settings import APISettings as _APISettings


USER_SETTINGS = getattr(settings, 'JWT_UTILS', None)

DEFAULTS = {
    # all settings with None or False are ignored by default.

    # Generating Jwt
    'SIGNING_KEY': settings.SECRET_KEY,
    'ALGORITHM': 'HS256',
    'ISSUER': None,  # None -> ignore (https://datatracker.ietf.org/doc/html/rfc7519#section-4.1.1)
    'SUBJECT': None,  # None -> ignore (https://datatracker.ietf.org/doc/html/rfc7519#section-4.1.2)
    'AUDIENCE': None,  # None -> ignore (https://datatracker.ietf.org/doc/html/rfc7519#section-4.1.3)
    'EXPIRATION': None,  # None -> ignore, ex: timedelta(minutes=5) (https://datatracker.ietf.org/doc/html/rfc7519#section-4.1.4)
    'NOT_BEFORE': None,  # None -> ignore, ex: timedelta(seconds=1) (https://datatracker.ietf.org/doc/html/rfc7519#section-4.1.5)
    'IAT': None,  # None -> ignore (https://datatracker.ietf.org/doc/html/rfc7519#section-4.1.6)
    'AUTO_IAT': False,  # Default False (generate iat with current time) - (used only if IAT is None)
    'JTI': None,  # None -> ignore (https://datatracker.ietf.org/doc/html/rfc7519#section-4.1.7)
    'AUTO_JTI': False,  # Default False (generate jti with random hex) - (used only if JTI is None)

    # Checking Jwt
    'VERIFY_ISSUER': False,  # False -> ignore
    'VERIFY_AUDIENCE': False,  # False -> ignore
    'VERIFY_EXPIRATION': False,  # False -> ignore
    'VERIFY_NOT_BEFORE': False,  # False -> ignore
    'VERIFY_IAT': False,  # False -> ignore
    'VERIFY_MAX_LEEWAY': 0,
}

DEFAULTS = {
    **DEFAULTS,
    'VERIFY_ISSUER': DEFAULTS['ISSUER'] is not None,
    'VERIFY_AUDIENCE': DEFAULTS['AUDIENCE'] is not None,
    'VERIFY_EXPIRATION': DEFAULTS['EXPIRATION'] is not None,
    'VERIFY_NOT_BEFORE': DEFAULTS['NOT_BEFORE'] is not None,
    'VERIFY_IAT': DEFAULTS['IAT'] is not None or DEFAULTS['AUTO_IAT'],
    'VERIFY_MAX_LEEWAY': 0,
}

IMPORT_STRINGS = ()

REMOVED_SETTINGS = ()


class APISettings(_APISettings):  # pragma: no cover
    pass


api_settings = APISettings(USER_SETTINGS, DEFAULTS, IMPORT_STRINGS)


def reload_api_settings(*args, **kwargs):  # pragma: no cover
    global api_settings

    setting, value = kwargs['setting'], kwargs['value']

    if setting == 'JWT_UTILS':
        api_settings = APISettings(value, DEFAULTS, IMPORT_STRINGS)


setting_changed.connect(reload_api_settings)
